{-
  The main module for the comment-line-count.
  This will print statistics about each module of the lines
  of code and the percentage of comments and haddock-documentation.
-}
module Main
  ( main )
where

{- Standard Library Modules Imported -}
import Control.Monad
  ( liftM )
import Data.List
  ( sortBy )
import System.Console.GetOpt
  ( getOpt
  , usageInfo
  , ArgOrder    ( .. )
  , OptDescr    ( .. )
  , ArgDescr    ( .. )
  )
import System.Directory
  ( getDirectoryContents
  , doesDirectoryExist
  )
import System.Environment
  ( getArgs )
import System.FilePath
  ( takeExtension
  , combine
  , splitDirectories
  )
{- External Library Modules Imported -}
{- Local Modules Imported -}
{- End of Imports -}

data CliFlag =
    CliHelp
  | CliShowFragments
  | CliSummarise
  | CliCommentPercent Int
  deriving Eq

          
happraiseOptions :: [ OptDescr CliFlag ]
happraiseOptions =
  [ Option   "h"     [ "help" ]
    (NoArg CliHelp)
    "Print the help message to standard out and then exit"

  , Option   "s"     [ "summarise" ]
    (NoArg CliSummarise)
    "Only print out the overall summary not each file's report"

  , Option   ""      [ "show-fragments" ]
    (NoArg CliShowFragments)
    "Actually print out the fragments of code and comment (debugging)"

  , Option ""        [ "comment-percent" ]
    (ReqArg (CliCommentPercent . read) "PERCENTAGE")
    "Say what percentage of each file should be comments to avoid a warning"
  ]

helpMessage :: String
helpMessage =
  unlines [ usageInfo "happraisal" happraiseOptions
          , "If your source code is under darcs a useful way to run happraise is:"
          , "happraise `darcs query manifest`"
          ]


-- | The main exported function 
main :: IO ()
main = getArgs >>= processOptions

processOptions :: [ String ] -> IO ()
processOptions cliArgs =
  case getOpt Permute  happraiseOptions cliArgs of
    (flags, args, [])       -> 
      processArgs flags args
    (_flags, _args, errors) -> 
      ioError $ userError (concat errors ++ helpMessage)


processArgs :: [ CliFlag ] -> [ String ] -> IO ()
processArgs flags files
  | elem CliHelp flags = putStrLn helpMessage
  | otherwise          =
    do fileReportLists <- mapM processFile files
       let fileReports = concat fileReportLists
       printOutput fileReports
  where
  printOutput :: [ FileReport ] -> IO ()
  printOutput fileReports
    | elem CliSummarise flags = putStrLn summary
    | otherwise               = putStrLn everything       
    where
    everything = unlines [ fileLines, summary ]
    fileLines  = unlines $ map (hprintFileReport flags) fileReports
    summary    = summarise flags fileReports


data FileReport =
  FileReport { frFile         :: FilePath
             , frLines        :: Int
             , frWords        :: Int
             , frCodeWords    :: Int
             , frCommentWords :: Int
             , frCommentRatio :: Double
             , frFragments    :: [ Fragment ]
             }

-- type Line = String
type Word = String

-- We return a list of file reports since it may be a directory
-- that we are processesing
processFile :: FilePath -> IO [ FileReport ]
processFile file =
  do isDir <- doesDirectoryExist file
     if isDir 
       then processDirectory file
       else if ".hs" == (takeExtension file)
            then do contents <- readFile file
                    return [ processContents file contents ]
            else return []

processDirectory :: FilePath -> IO [ FileReport ]
processDirectory file =
  do filesUnF <- getDirectoryContents file
     let files = map (combine file) $ filter fileFilter filesUnF
     liftM concat $ mapM processFile files
  where
  fileFilter :: FilePath -> Bool
  fileFilter f = not (isCurOrAbove f || isDarcsDir f)


  -- Checks if it is a _darcs directory which we do not bother to
  -- process since it will contain (for example) a pristine version
  -- of the source code tree, and we will just be going over the same
  -- files twice.
  isDarcsDir :: FilePath -> Bool
  isDarcsDir "_darcs"  = True
  isDarcsDir "_darcs/" = True
  isDarcsDir f         = "_darcs" == (last $ splitDirectories f)

  -- Checks if the given directory is the current or the directory above
  -- this avoids us going into an infinite recursive loop.
  isCurOrAbove :: FilePath -> Bool
  isCurOrAbove "./"  = True
  isCurOrAbove "../" = True
  isCurOrAbove "."   = True
  isCurOrAbove ".."  = True
  isCurOrAbove _     = False

hprintFileReport :: [ CliFlag ] -> FileReport -> String
hprintFileReport flags fr =
  unlines everything
  where
  everything = concat [ header, lowCommentWarning, stock, showSeparation ]
  header :: [ String ]
  header = [ "Report for the file: " ++ (frFile fr) ]
  -- ie the stuff that's always printed
  stock :: [ String ]
  stock = [ declare "   Number of source lines: "  $ frLines fr
          , declare "   Number of words : "        $ frWords fr
          , declare "   Number of code words : "   $ codeWords
          , declare "   Number of comment words: " $ commentWords
          , declare "   The comment ratio is: "    $ frCommentRatio fr
          ]
          
  lowCommentWarning :: [ String ]
  lowCommentWarning
    | lowCommentCondition flags fr = [ "   ---------------------------"
                                     , "   WARNING: low comment"
                                     , "   ---------------------------"
                                     ]
    | otherwise                    = []

  codeWords    = frCodeWords fr
  commentWords = frCommentWords fr

  declare :: Show a => String -> a -> String
  declare s a = s ++ (show a)

  showSeparation :: [ String ]
  showSeparation 
    | elem CliShowFragments flags =
      [ "-----------The Code ---------"
      , showCode $ frFragments fr
      , "-----------The Comments ------"
      , showComment $ frFragments fr
      ]
    | otherwise                   =
      []

  showCode :: [ Fragment ] -> String
  showCode = unwords . (map fragmentString) . (filter isCode)

  showComment :: [ Fragment ] -> String
  showComment = unwords . (map fragmentString) . (filter isComment)

  fragmentString :: Fragment -> String
  fragmentString (Code s)    = unwords s
  fragmentString (Comment s) = unwords s

  isCode :: Fragment -> Bool
  isCode (Code _) = True
  isCode _        = False
    
  isComment :: Fragment -> Bool
  isComment (Comment _) = True
  isComment _           = False

processContents :: FilePath -> String -> FileReport
processContents file contents =
  FileReport { frFile         = file
             , frLines        = noLines
             , frWords        = noWords
             , frCodeWords    = noCode
             , frCommentWords = noComment
             , frCommentRatio = commentRatio
             , frFragments    = fragments
             }
  where
  cLines    = lines contents
  noLines   = length cLines
  
  wordLines = map words cLines
  noWords   = foldr1 (+) $ map length wordLines

  fragments = categoriseWords contents
  
  noCode    = foldr1 (+) $ map codeWordsFragment fragments
  noComment = foldr1 (+) $ map commentWordsFragment fragments

  codeWordsFragment :: Fragment -> Int
  codeWordsFragment (Code ws)   = length ws
  codeWordsFragment (Comment _) = 0

  commentWordsFragment :: Fragment -> Int
  commentWordsFragment (Comment ws) = length ws
  commentWordsFragment (Code _)     = 0
  

  commentRatio =  (fromIntegral $ noComment) / (fromIntegral $ noWords)


data Fragment =
    Code       [ Word ]
  | Comment    [ Word ]
  deriving Show
  -- DocComment [ Word ]

{-
  categorise the words in to fragments
-}
categoriseWords :: String -> [ Fragment ]
categoriseWords []                 = []
categoriseWords ('{' : '-' : rest) =
  (Comment $ words comment) : (categoriseWords remainder)
  where
  (comment, remainder) = getNestedComment "-{" rest
categoriseWords l@('-' : '-' : _)  =
  (Comment $ words comment) : (categoriseWords remainder)
  where
  (comment, remainder) = getLineComment l
categoriseWords ('"' : rest)       =
  (Code $ words string) : (categoriseWords remainder)
  where
  (string, remainder) = codeUpStringLiteral "\"" rest
categoriseWords l                  =
  (Code $ words code) : (categoriseWords remainder)
  where
  (code, remainder) = skipToComment "" l

getNestedComment :: String -> String -> (String, String)
getNestedComment soFar ('-' : '}' : rest) =
  (reverse $ "}-" ++ soFar, rest)
getNestedComment soFar ('{' : '-' : rest) =
  getNestedComment (innerComment ++ soFar) remainder
  where
  (innerComment, remainder) = getNestedComment "-{" rest
getNestedComment soFar (h : rest)         =
  getNestedComment (h : soFar) rest
getNestedComment soFar []                 =
  (reverse soFar, [])

getLineComment :: String -> (String, String)
getLineComment = break (== '\n')
    
skipToComment :: String -> String -> (String, String)
skipToComment soFar l@('{' : '-' : _)   =
  (reverse soFar, l)
skipToComment soFar l@('-' : '-' : _)   =
  (reverse soFar, l)
skipToComment soFar []                  =
  (reverse soFar, [])
skipToComment soFar ('\'' : '"' : rest) =
  skipToComment ('\'' : '"' : soFar) rest
skipToComment soFar l@('"' : _)         =
  (reverse soFar, l)
skipToComment soFar (h : rest)          =
  skipToComment (h : soFar) rest

codeUpStringLiteral :: String -> String -> ( String, String )
codeUpStringLiteral soFar ('"' : rest)        =
  (reverse $ '"' : soFar, rest)
codeUpStringLiteral soFar ('\\' : '"' : rest) =
  codeUpStringLiteral ('\\' : '"' : soFar) rest
codeUpStringLiteral soFar (h : rest)          =
  codeUpStringLiteral (h : soFar) rest
codeUpStringLiteral soFar ""                  =
  (reverse soFar, "")


{-
  Summarising a list of file reports the idea here is to present a
  less verbose summary of the file reports.
  Eventually someone might run:
    $ happraise src
  and if they get back:
    summary : 0 files had too few comments
  then that's fine but if they get back:
    summary : 4 files had too few comments
  then they can go and look at those individual files and try to fix them.
-}
summarise :: [ CliFlag ] -> [ FileReport ] -> String
summarise flags fileReports =
  unlines $ header : fileLines
  where
  header = case length tooFewComments of
            0  -> "summary : everything is a-okay"
            1  -> "1 file had too few comments"
            x  -> (show x) ++ " files had too few comments"
  tooFewComments = sortOnCommentPercent $ filter condition fileReports

  fileLines      = map makeFileLine tooFewComments

  -- Our first condition is simply whether or not the number
  -- of comment words is larger than the number of code words.
  condition :: FileReport -> Bool
  condition = lowCommentCondition flags

  -- A file line is just the file name indented.
  makeFileLine :: FileReport -> String
  makeFileLine fr = concat [ "    "
                           , frFile fr
                           , " : "
                           , show $ frCommentRatio fr
                           ]


  sortOnCommentPercent :: [ FileReport ] -> [ FileReport ]
  sortOnCommentPercent = sortBy compareCommentPercent

  compareCommentPercent :: FileReport -> FileReport -> Ordering
  compareCommentPercent fr1 fr2 =
    compare (frCommentRatio fr1) (frCommentRatio fr2)

-- Should return true if there is a LOW comment ratio
-- that is IF there should be a warning issued.
lowCommentCondition :: [ CliFlag ] -> FileReport -> Bool
lowCommentCondition flags fr
    -- If no percentage is given the default is 50 percent
  | null percents       = (frCommentRatio fr) <= 0.5
    -- If one is given then we just take the first one given
    -- perhaps we could warn/error if more than one is given.
  | otherwise           = 
    (frCommentRatio fr) <= ( percent / 100.0 )
  where
  percents = [ p  | CliCommentPercent p <- flags ]

  -- We of course only use this if 'percents' is not null
  -- see the pattern guard above.
  percent :: Double
  percent  = fromIntegral $ head percents
